// keyb.c
//
// TV Remote controller project for Motorola68HC908QY4
// Based on SDCC M08 Port Compiler & Library
// By Sarosi, 2004.03.07, Free
//
#include "mc68hc908qy.h"
#include "rc.h"
#include "delay.h"

/*
Keyboard:
                       5V   5V   5V   5V
                        |    |    |    |
                        _    _    _    _
     ----------+       | |  | |  | |  | |
               |       |_|  |_|  |_|  |_|
               |        |    |    |    |
     PTA2/KBI2 <--------o    |    |    |
               |        |    |    |    |
     PTA3/KBI2 <-------------o    |    |
               |        |    |    |    |
     PTA5/KBI2 <--------|----|----o    |
               |        |    |    |    |
     PTA6/KBI2 <--------|----|----|----o
               |        |    |    |    |
               | 0V     |    |    |    |
          PTB0 >---|<|--X----X----X----X--
               |        |    |    |    |
          PTB1 >---|<|--X----X----X----X--
               |        |    |    |    |
          PTB2 >---|<|--X----X----X----X--
               |        |    |    |    |
          PTB3 >---|<|--X----X----X----X--
               |        |    |    |    |
     ----------+


Key layout:

        1  2  3        XA = X D
        4  5  6        XB = X Y
        7  8  9        XC = Y 1
           0           XD = Y 2
        L  U  R
        X  D  Y
*/


unsigned char _near key; // Key by IT


//-----------------------------------------------
// KeyInit - init keyboard matrix (IT disabled, but pollable)
//-----------------------------------------------
void KeyInit(void){

  // set pins to input (DDRA2 is input-only)
  DDRA3=0;
  DDRA4=0;
  DDRA5=0;

  // set key matrix outputs to low
  PTB  &= 0xF0;
  DDRB |= 0x0F;

  // prepare keyboard logic
  IMASKK = 1;    // disable IT
  KBIER |= 0x3C; // enable KBIE2..5
  MODEK  = 1;    // IT on low level
  ACKK   = 1;    // Clear pending status
  ACKK   = 0;    
  key    = 0;
  //IMASKK = 0;  // enable IT
}


//-----------------------------------------------
// KeyInRow - gets key code [1...4] for actual row, or 0 if not
//-----------------------------------------------
unsigned char KeyInRow(unsigned char base){
  unsigned char rowBits, col, key;

  for(key=0, rowBits = PTA, col=1;  col!=5;  col++){
     if( ! (rowBits & 4) ){
        if( key ) return 0; // more then one
        key = col;
     }
     rowBits>>=1;
  }
  if( key ) key += base;
  return key;
}


//-----------------------------------------------
// KeyInMatrix - reads the actual compound key from matrix
//-----------------------------------------------
unsigned char KeyInMatrix(void){
  unsigned char key0, key1, key2, key3;

  // scan rows one by one
  PTB0=0; PTB1=1; PTB2=1; PTB3=1;
  key0 = KeyInRow(0);  // [1,2,3,4]

  PTB0=1; PTB1=0;
  key1 = KeyInRow(4);  // [4,6,7,8]

  PTB1=1; PTB2=0;
  key2 = KeyInRow(8);  // [9,10,11,12]

  PTB2=1; PTB3=0;
  key3 = KeyInRow(12);  // [13,14,15,16]

  // restore key matrix outputs
  PTB0=0; PTB1=0; PTB2=0; PTB3=0;

  // scan compaund key presses
  if( key0 ){
     if( key1 ){
       return 17;
     } else if( key2 ){
       return 18;
     } else if( key3 ){
       return 19;
     } else {
       return key0;
     }
  } else if( key1){
     if( key2 ){
       return 20;
     } else if( key3 ){
       return 21;
     } else {
       return key1;
     }
  } else if( key2 ){
     if( key3 ){
       return 22;
     } else {
       return key2;
     }
  } else if( key3 ){
     return key3;
  }

  return 0;
}


//-----------------------------------------------
// GetKey - reads in a key
//-----------------------------------------------
unsigned char GetKey(void){

// Key handling:
//   - Key is returned when all keys are released
//   - Compound key (code>16) has higher priority as noncompound keys

unsigned char keyT_2, keyT_1, keyT_0;  // Key scans
unsigned char pressedKey;  // Time[ms]
#define TIME_SLICE 10 //ms

   // init states
   keyT_2 = 0;
   keyT_1 = 0;
   keyT_0 = KeyInMatrix();

   pressedKey = 0;

   // scan matrix in time slices
   while( keyT_2 || keyT_1 || keyT_0 ){

     // slice
     DelayMs( TIME_SLICE );

     // shift states, and scan matrix
     keyT_2 = keyT_1;
     keyT_1 = keyT_0;
     keyT_0 = KeyInMatrix();

     if(  keyT_0 && keyT_0==keyT_1 && keyT_1==keyT_2 ){
        // a key is firmly pressed
        if( pressedKey <= 16 ){
           pressedKey = keyT_2;
        }
     }
   }

   // all released
   return pressedKey;
}


//-----------------------------------------------
// Keyboard_IT - IT vector still must be planted anyway
//-----------------------------------------------
void Keyboard_IT(void) interrupt 0xffE0 {
    key = GetKey();
    IMASKK = 1;   // disable IT

    _asm
    .area _ABSOLUTE(ABS)
    .org 0xffE0
    .dw _Keyboard_IT
    .area _CODE
    _endasm;
}

// End of keyb.c
